# -*- coding: utf8 -*-

"""

   Copyright (C) 2011-2025 Andreas Würl

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""

from __future__ import print_function

import datetime
import sys
from zoneinfo import ZoneInfo

import math
import shapely.wkt
from optparse import OptionParser

import blitzortung.config
import blitzortung.db.query
import blitzortung.geom
import blitzortung.util

DEFAULT_GRID = 1.0, 1.0

DATE_FORMAT = '%Y%m%d'
TIME_FORMAT = '%H%M'
SECONDS_FORMAT = '%S'


def parse_time(date_string, time_string, tz, description, is_end_time=False):
    try:
        has_seconds = len(time_string) > 4
        tmp_date = datetime.datetime.strptime(date_string, DATE_FORMAT).date()
        tmp_time = datetime.datetime.strptime(time_string,
                                              TIME_FORMAT + (SECONDS_FORMAT if has_seconds else '')).time()
        parsed_time = datetime.datetime.combine(tmp_date, tmp_time).replace(tzinfo=tz)
        if is_end_time:
            parsed_time += (datetime.timedelta(seconds=1) if has_seconds else datetime.timedelta(minutes=1))
        return parsed_time
    except Exception:
        print("parse error in %s: '%s %s'" % description, date_string, time_string)
        sys.exit(5)


def prepare_grid_if_applicable(options, area):
    if options.grid is not None or options.xgrid is not None or options.ygrid is not None:

        grid_x, grid_y = DEFAULT_GRID

        if options.grid is not None:
            grid_x, grid_y = options.grid, options.grid

        if options.xgrid is not None:
            grid_x = options.xgrid
            if options.ygrid is None and options.grid is None:
                grid_y = grid_x

        if options.ygrid is not None:
            grid_y = options.ygrid
            if options.xgrid is None and options.grid is None:
                grid_x = grid_y

        if options.area is None:
            print("grid options requires declaration of envelope area")
            sys.exit(1)

        env = area.envelope.bounds
        return blitzortung.geom.Grid(env[0], env[2], env[1], env[3], grid_x, grid_y, options.srid)


def main():
    end_time = datetime.datetime.now(datetime.UTC)
    start_time = end_time - datetime.timedelta(hours=1)
    end_time -= datetime.timedelta(minutes=1)

    parser = OptionParser()

    parser.add_option("--startdate", dest="startdate", default="default",
                      help="start date for data retrieval", type="string")

    parser.add_option("--starttime", dest="starttime", default="default",
                      help="start time for data retrieval", type="string")

    parser.add_option("--enddate", dest="enddate", default="default",
                      help="end date for data retrieval", type="string")

    parser.add_option("--endtime", dest="endtime", default="default",
                      help="end time for data retrieval", type="string")

    parser.add_option("--area", dest="area",
                      help="area for which strikes are selected", type="string")

    parser.add_option("--tz", dest="tz", default=str(datetime.UTC),
                      help="used timezone", type="string")

    parser.add_option("--useenv", dest="useenv", default=False, action="store_true",
                      help="use envelope of given area for query")

    parser.add_option("--srid", dest="srid", default=blitzortung.geom.Geometry.default_srid,
                      help="srid for query area and results", type="int")

    parser.add_option("--precision", dest="precision", default=4,
                      help="precision of coordinates", type="int")

    parser.add_option("--grid", dest="grid",
                      help="grid width", type="float")

    parser.add_option("--x-grid", dest="xgrid",
                      help="grid x width", type="float")

    parser.add_option("--y-grid", dest="ygrid",
                      help="grid y width", type="float")

    parser.add_option("--map", dest="map", action="store_true",
                      help="show ascii map instead of numerical grid")

    (options, args) = parser.parse_args()

    try:
        tz = ZoneInfo(options.tz)
    except:
        print('parse error in timezone "' + options.tz + '"')
        sys.exit(1)

    start_time = start_time.astimezone(tz)
    end_time = end_time.astimezone(tz)

    if options.startdate == 'default':
        options.startdate = start_time.strftime(DATE_FORMAT)
    if options.starttime == 'default':
        options.starttime = start_time.strftime(TIME_FORMAT)
    non_default_end = options.enddate != 'default' or options.endtime != 'default'
    if options.enddate == 'default':
        options.enddate = end_time.strftime(DATE_FORMAT)
    if options.endtime == 'default':
        options.endtime = end_time.strftime(TIME_FORMAT)

    start_time = parse_time(options.startdate, options.starttime, tz, "starttime")
    end_time = parse_time(options.enddate, options.endtime, tz, "endtime",
                          is_end_time=True) if non_default_end else None

    area = None
    if options.area:
        try:
            area = shapely.wkt.loads(options.area)
        except shapely.errors.GEOSException:
            print('parse error in area "' + options.area + '"')

        if area and options.useenv:
            area = area.envelope

    # open strike database
    strike_db = blitzortung.db.strike()

    # set data parameters
    strike_db.set_srid(options.srid)
    strike_db.set_timezone(tz)

    # set time range
    time_interval = blitzortung.db.query.TimeInterval(start_time, end_time)

    # set query parameters
    order = 'timestamp'

    grid = prepare_grid_if_applicable(options, area)

    # start timer for database select
    timer = blitzortung.util.Timer()

    if grid:
        grid_result = strike_db.select_grid(grid, time_interval=time_interval)

        select_time = timer.lap()

        print(grid_result.to_map() if options.map else grid_result.to_arcgrid())
        sys.stderr.write('received grid data in %.3f seconds\n' % (select_time))

    else:
        strikes = strike_db.select(time_interval=time_interval, geometry=area, order=order)

        precision_factor = math.pow(10.0, options.precision)

        strike_count = 0
        for strike in strikes:
            strike_count += 1
            strike.x = round(strike.x * precision_factor) / precision_factor
            strike.y = round(strike.y * precision_factor) / precision_factor
            print(str(strike))

        select_time = timer.lap()

        sys.stderr.write('received %d strikes in %.3f seconds\n' % (strike_count, select_time))


if __name__ == '__main__':
    main()



# -*- coding: utf8 -*-

"""

   Copyright 2014-2016 Andreas Würl

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""

import time

from injector import inject

from .db import execute
from .. import db
from ..db.query import TimeInterval


class HistogramQuery:
    @inject
    def __init__(self, strike_query_builder: db.query_builder.Strike):
        self.strike_query_builder = strike_query_builder

    def create(self, time_interval: TimeInterval, connection_pool, region=None, envelope=None):
        reference_time = time.time()

        query = self.strike_query_builder.histogram_query(db.table.Strike.table_name, time_interval, 5, region,
                                                          envelope)

        result = execute(connection_pool, query)
        result.addCallback(self.build_result, minutes=time_interval.minutes(), bin_size=5,
                           reference_time=reference_time)
        return result

    @staticmethod
    def build_result(query_result, minutes, bin_size, reference_time):
        time_duration = time.time() - reference_time
        print("histogram: query %.03fs" % time_duration)
        value_count = int(minutes / bin_size)

        result = [0] * value_count

        for bin_data in query_result:
            result[bin_data[0] + value_count - 1] = bin_data[1]

        return result



# -*- coding: utf8 -*-

"""

   Copyright 2014-2016 Andreas Würl

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""
import datetime
import logging

import math
from typing import Optional

from blitzortung.db.grid_result import build_grid_result
from injector import inject

from . import mapper
from . import query
from . import query_builder
from .query import TimeInterval
from .. import data
from .. import geom
from ..logger import get_logger_name

import psycopg2
import psycopg2.pool
import psycopg2.extras
import psycopg2.extensions

from abc import ABCMeta, abstractmethod


class Base:
    """
    abstract base class for database access objects

    creation of database

    as user postgres:

    createuser -i -D -R -S -W -E -P blitzortung
    createdb -E utf8 -O blitzortung blitzortung
    createlang plpgsql blitzortung
    psql -f /usr/share/postgresql/10/contrib/postgis-2.4/postgis.sql -d blitzortung
    psql -f /usr/share/postgresql/12/contrib/postgis-3.0/postgis.sql -d blitzortung
    psql -f /usr/share/postgresql/10/contrib/postgis-2.4/spatial_ref_sys.sql -d blitzortung

    psql blitzortung

    GRANT SELECT ON spatial_ref_sys TO blitzortung;
    GRANT SELECT ON geometry_columns TO blitzortung;
    GRANT INSERT, DELETE ON geometry_columns TO blitzortung;
    CREATE EXTENSION "btree_gist";

    """
    __metaclass__ = ABCMeta

    default_timezone = datetime.timezone.utc

    def __init__(self, db_connection_pool):

        self.logger = logging.getLogger(get_logger_name(self.__class__))
        self.db_connection_pool = db_connection_pool

        self.schema_name = ""
        self.table_name = ""

        while True:
            self.conn = self.db_connection_pool.getconn()
            self.conn.cancel()
            try:
                self.conn.reset()
            except psycopg2.OperationalError:
                print("reconnect to db")
                self.db_connection_pool.putconn(self.conn, close=True)
                continue
            break

        psycopg2.extensions.register_type(psycopg2.extensions.UNICODE, self.conn)
        psycopg2.extensions.register_type(psycopg2.extensions.UNICODEARRAY, self.conn)
        self.conn.set_client_encoding('UTF8')

        self.srid = geom.Geometry.default_srid
        self.tz = None
        self.set_timezone(Base.default_timezone)

        cur = None
        try:
            cur = self.conn.cursor(cursor_factory=psycopg2.extras.DictCursor)
        except psycopg2.DatabaseError as error:
            self.logger.error(error)

            if self.conn:
                try:
                    self.conn.close()
                except NameError:
                    pass
        finally:
            if cur:
                cur.close()

    def close(self):
        self.db_connection_pool.putconn(self.conn)

    def is_connected(self):
        if self.conn:
            return not self.conn.closed
        else:
            return False

    @property
    def full_table_name(self):
        if self.schema_name:
            return '"' + self.schema_name + '"."' + self.table_name + '"'
        else:
            return self.table_name

    def get_srid(self):
        return self.srid

    def set_srid(self, srid):
        self.srid = srid

    def get_timezone(self):
        return self.tz

    def set_timezone(self, tz):
        self.tz = tz
        with self.conn.cursor() as cur:
            cur.execute('SET TIME ZONE \'%s\'' % str(self.tz))

    def fix_timezone(self, timestamp):
        return timestamp.astimezone(self.tz) if timestamp else None

    def from_bare_utc_to_timezone(self, utc_time):
        return utc_time.replace(tzinfo=datetime.timezone.utc).astimezone(self.tz)

    @staticmethod
    def from_timezone_to_bare_utc(time_with_tz):
        return time_with_tz.astimezone(datetime.timezone.utc).replace(tzinfo=None)

    def commit(self):
        """ commit pending database transaction """
        self.conn.commit()

    def rollback(self):
        """ rollback pending database transaction """
        self.conn.rollback()

    @abstractmethod
    def insert(self, *args):
        pass

    @abstractmethod
    def select(self, **kwargs):
        pass

    def execute(self, sql_statement, parameters=None, factory_method=None, **factory_method_args):
        with self.conn.cursor(cursor_factory=psycopg2.extras.DictCursor) as cursor:
            cursor.execute(sql_statement, parameters)
            if factory_method:
                method = factory_method(cursor, **factory_method_args)
                return method

    def execute_single(self, sql_statement, parameters=None, factory_method=None, **factory_method_args):
        def single_cursor_factory(cursor):
            if cursor.rowcount == 1:
                return factory_method(cursor.fetchone(), **factory_method_args)

        return self.execute(sql_statement, parameters, single_cursor_factory)

    def execute_many(self, sql_statement, parameters=None, factory_method=None, **factory_method_args):
        factory_method = factory_method or (lambda values, **_: values)
        with self.conn.cursor(cursor_factory=psycopg2.extras.DictCursor) as cursor:
            cursor.execute(sql_statement, parameters)
            for value in cursor:
                yield factory_method(value, **factory_method_args)


class Strike(Base):
    """
    strike db access class

    database table creation (as db user blitzortung, database blitzortung):

    > psql -h localhost -U blitzortung -W blitzortung

    CREATE TABLE strikes (id bigserial, "timestamp" timestamptz, nanoseconds SMALLINT, geog GEOGRAPHY(Point),
        PRIMARY KEY(id));
    ALTER TABLE strikes ADD COLUMN altitude SMALLINT;
    ALTER TABLE strikes ADD COLUMN region SMALLINT;
    ALTER TABLE strikes ADD COLUMN amplitude REAL;
    ALTER TABLE strikes ADD COLUMN error2d SMALLINT;
    ALTER TABLE strikes ADD COLUMN stationcount SMALLINT;

    CREATE INDEX strikes_timestamp ON strikes USING btree("timestamp");
    CREATE INDEX strikes_timestamp_geog ON strikes USING gist("timestamp", geog);
    CREATE INDEX strikes_region_timestamp ON strikes USING btree(region, "timestamp");
    CREATE INDEX strikes_id_timestamp ON strikes USING btree(id, "timestamp");
    CREATE INDEX strikes_geog ON strikes USING gist(geog);
    CREATE INDEX strikes_id_timestamp_geog ON strikes USING gist(id, "timestamp", geog);
    CREATE INDEX strikes_region_timestamp_nanoseconds ON strikes USING btree(region, "timestamp", nanoseconds);

    empty the table with the following commands:

    DELETE FROM strikes;
    ALTER SEQUENCE strikes_id_seq RESTART 1;

    """

    table_name = 'strikes'

    @inject
    def __init__(self, db_connection_pool: psycopg2.pool.ThreadedConnectionPool, query_builder_: query_builder.Strike,
                 strike_mapper: mapper.Strike):
        super().__init__(db_connection_pool)

        self.table_name = Strike.table_name
        self.query_builder = query_builder_
        self.strike_mapper = strike_mapper

    def insert(self, strike, region=1):
        sql = 'INSERT INTO ' + self.full_table_name + \
              ' ("timestamp", nanoseconds, geog, altitude, region, amplitude, error2d, stationcount) ' + \
              'VALUES (%(timestamp)s, %(nanoseconds)s, ST_MakePoint(%(longitude)s, %(latitude)s), ' + \
              '%(altitude)s, %(region)s, %(amplitude)s, %(error2d)s, %(stationcount)s)'

        parameters = {
            'timestamp': strike.timestamp.datetime,
            'nanoseconds': strike.timestamp.nanosecond,
            'longitude': strike.x,
            'latitude': strike.y,
            'altitude': strike.altitude,
            'region': region,
            'amplitude': strike.amplitude,
            'error2d': strike.lateral_error,
            'stationcount': strike.station_count
        }

        self.execute(sql, parameters)

    def get_latest_time(self, region=None):
        sql = 'SELECT "timestamp", nanoseconds FROM ' + self.full_table_name + \
              (' WHERE region=%(region)s' if region else '') + \
              ' ORDER BY "timestamp" DESC, nanoseconds DESC LIMIT 1'

        def prepare_result(result):
            return data.Timestamp(self.fix_timezone(result['timestamp']), result['nanoseconds'])

        parameters = {'region': region}
        return self.execute_single(sql, parameters, prepare_result)

    def select(self, **kwargs):
        """ build up query """

        query_ = self.query_builder.select_query(self.full_table_name, self.srid, **kwargs)

        return self.execute_many(str(query_), query_.get_parameters(), self.strike_mapper.create_object,
                                 timezone=self.tz)

    def select_grid(self, grid, count_threshold, **kwargs):
        """ build up raster query """

        query = self.query_builder.grid_query(self.table_name, grid, count_threshold, **kwargs)
        data = self.execute_many(str(query), query.get_parameters())

        grid_result = build_grid_result(data, grid.x_bin_count, grid.y_bin_count, kwargs['time_interval'].end)

        return grid_result

    def select_global_grid(self, grid, count_threshold, **kwargs):
        """ build up raster query """

        query = self.query_builder.global_grid_query(self.table_name, grid, count_threshold, **kwargs)

        data = self.execute_many(str(query), query.get_parameters())

        grid_result = build_grid_result(data, grid.x_bin_count, grid.y_bin_count, kwargs['time_interval'].end)

        return grid_result

    def select_histogram(self, time_interval: TimeInterval, binsize: int = 5, region: Optional[int] = None,
                         envelope=None):

        query = self.query_builder.histogram_query( self.full_table_name, time_interval, binsize, region, envelope )

        minutes = time_interval.minutes()

        def prepare_result(cursor):
            value_count = minutes // binsize

            result = [0] * value_count

            for bin_data in cursor:
                result[bin_data[0] + value_count - 1] = bin_data[1]

            return result

        return self.execute(str(query), query.get_parameters(), prepare_result)



import os.path
import sys

from twisted.scripts.twistd import run


def main():
    target_dir = os.path.dirname(os.path.abspath(__file__))

    args = ["twistd"]
    if not os.environ.get("BLITZORTUNG_TEST"):
        args += ["--pidfile", "/var/run/bo-webservice.pid"]

    sys.argv = args + ["-oy", os.path.join(target_dir, "webservice.py")]
    sys.exit(run())


if __name__ == "__main__":
    main()



# -*- coding: utf8 -*-

"""
   Copyright (C) 2011-2025 Andreas Würl

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""

import datetime
import logging
import os
from contextlib import nullcontext

import requests
import statsd
from optparse import OptionParser

import blitzortung.config
import blitzortung.db
import blitzortung.db.query
import blitzortung.logger
from blitzortung import util
from blitzortung.data import Timestamp
from blitzortung.lock import LockWithTimeout, FailedToAcquireException

logger = logging.getLogger(os.path.basename(__file__))
blitzortung.set_parent_logger(logger)
blitzortung.add_log_handler(blitzortung.logger.create_console_handler())

statsd_client = statsd.StatsClient('localhost', 8125, prefix='org.blitzortung.import')


def fetch_strikes_from_url(url, auth=None):
    """
    Fetch strike data from a given URL and parse it into Strike objects.

    The URL returns JSON-formatted strike data, one strike per line.
    Each line has the format:
    {"time":1763202124325980200,"lat":-15.296556,"lon":134.589548,"alt":0,"pol":0,...}

    Args:
        url: The URL to fetch strike data from
        auth: Optional tuple of (username, password) for authentication

    Yields:
        Strike objects parsed from the URL response
    """
    import json
    from blitzortung.builder import Strike as StrikeBuilder


    try:
        timer = util.Timer()
        response = requests.get(url, auth=auth, timeout=30)
        response.raise_for_status()

        logger.info("Fetching strikes from URL: %s (%.03fs)", url, timer.lap())

        builder = StrikeBuilder()

        strike_count = 0

        for line in response.text.splitlines():
            line = line.strip()
            if not line:
                continue

            try:
                # Parse JSON data
                data = json.loads(line)

                # Create strike from JSON data
                # Build strike object (create new builder for each strike)
                strike = (builder
                          .set_timestamp(Timestamp(data['time']))
                          .set_x(round(data['lon'],4))
                          .set_y(round(data['lat'],4))
                          .set_altitude(data.get('alt', 0))
                          .set_amplitude(data.get('pol', 0))
                          .set_lateral_error(data.get('mds', 0))
                          .build())

                strike_count += 1
                yield strike

            except (json.JSONDecodeError, KeyError) as e:
                logger.warning("Failed to parse strike: %s (%s)", e, line)
                continue
            except Exception as e:
                logger.warning("Failed to create strike object: %s (%s)", e, line)
                continue

        logger.info("Fetched %d strikes from URL", strike_count)

    except requests.RequestException as e:
        logger.error("Failed to fetch data from URL %s: %s", url, e)
        raise


def create_strike_key(strike):
    """
    Create a unique key for a strike based on its attributes.

    Since strikes from URLs don't have IDs, we identify them by:
    - timestamp (nanosecond precision)
    - location (x, y coordinates)
    - lateral error

    Args:
        strike: Strike object

    Returns:
        Tuple representing the strike's unique characteristics
    """
    return (
        strike.timestamp.value,
        round(strike.x, 4),  # Round to 4 decimal places for location
        round(strike.y, 4),
        strike.lateral_error
    )


def get_existing_strike_keys(strike_db, time_interval):
    """
    Retrieve keys of strikes already present in the database for a given time interval.

    Strikes are identified by their timestamp, location, and amplitude since
    strikes from URLs don't have database IDs.

    Args:
        strike_db: Database connection for strikes
        time_interval: Time interval to query

    Returns:
        Set of strike keys (tuples of timestamp, x, y, amplitude)
    """
    logger.debug("Querying existing strikes for interval %s - %s",
                 time_interval.start, time_interval.end)

    existing_strikes = strike_db.select(time_interval=time_interval, order="timestamp")
    strike_keys = {create_strike_key(strike) for strike in existing_strikes}

    logger.info("Found %d existing strikes in database", len(strike_keys))
    return strike_keys


def update_strikes(hours=1):
    """
    Update strike database by fetching data from a URL and inserting new strikes.

    This function:
    1. Calculates a time interval (default: last 1 hour)
    2. Retrieves existing strikes from the database for that interval
    3. Fetches strikes from the provided URL
    4. Inserts only strikes that are not already in the database

    Args:
        url: URL to fetch strike data from (if None, uses default config URL)
        hours: Number of hours to look back (default: 1)

    Returns:
        Number of new strikes inserted into the database
    """
    logger.info("Starting strike update (looking back %d hour(s))", hours)

    now = datetime.datetime.now(datetime.timezone.utc)
    start_time = now - datetime.timedelta(hours=hours)

    # Get configuration if URL not provided
    config = blitzortung.config.config()
    start_timestamp_ns = int(start_time.timestamp() * 1e6) * 1000
    url = f"https://data.blitzortung.org/Data/Protected/last_strikes.php?time={start_timestamp_ns}"
    auth = (config.get_username(), config.get_password())

    # Calculate time interval (last N hours)
    end_time = now
    time_interval = blitzortung.db.query.TimeInterval(
        start_time,
        end_time
    )

    logger.info("Time interval: %s to %s", start_time, end_time)

    # Get database connection
    strike_db = blitzortung.db.strike()

    # Get existing strikes from database (identified by timestamp/location/amplitude)
    existing_strike_keys = get_existing_strike_keys(strike_db, time_interval)
    for existing_strike_key in existing_strike_keys:
        logger.debug("Existing strike key: %s", existing_strike_key)

    # Fetch strikes from URL
    try:
        url_strikes = list(fetch_strikes_from_url(url, auth=auth))
    except requests.RequestException as e:
        logger.error("Failed to fetch strikes from URL: %s", e)
        return 0

    # Filter strikes: only those within time interval and not in database
    new_strikes = []
    for strike in url_strikes:
        # Check if strike is within the time interval
        if not (time_interval.start <= strike.timestamp <= time_interval.end):
            logger.debug("Strike at %s outside time interval, skipping", strike.timestamp)
            continue

        # Check if strike already exists in database (by timestamp/location/amplitude)
        strike_key = create_strike_key(strike)
        if strike_key in existing_strike_keys:
            logger.debug("Strike at %s (%.4f, %.64) already exists, skipping",
                        strike.timestamp, strike.x, strike.y)
            continue

        new_strikes.append(strike)

    logger.info("Found %d new strikes to insert (out of %d from URL)",
                len(new_strikes), len(url_strikes))

    # Insert new strikes
    insert_count = 0
    for strike in new_strikes:
        try:
            strike_db.insert(strike)
            insert_count += 1

        except Exception as e:
            logger.error("Failed to insert strike %s: %s", strike.id, e)
            strike_db.rollback()
            raise

    # Final commit
    if insert_count > 0:
        strike_db.commit()
        logger.info("Successfully inserted %d new strikes", insert_count)
    else:
        logger.info("No new strikes to insert")

    strike_db.close()

    # Update statistics
    statsd_client.gauge("strikes.imported", insert_count)

    return insert_count





def main():
    """
    Command-line interface for the strike import tool.
    """
    parser = OptionParser(description="Import strike data from URL into database")
    parser.add_option("--hours", dest="hours", type="int", default=1,
                      help="Number of hours to look back (default: 1)")
    parser.add_option("-v", "--verbose", dest="verbose", action="store_true",
                      help="Enable verbose logging")
    parser.add_option("-d", "--debug", dest="debug", action="store_true",
                      help="Enable debug logging")
    parser.add_option("--no-lock", dest="no_lock", action="store_true",
                      help="Skip file locking (use with caution)")

    (options, args) = parser.parse_args()

    # Set logging level
    if options.debug:
        blitzortung.set_log_level(logging.DEBUG)
    elif options.verbose:
        blitzortung.set_log_level(logging.INFO)
    else:
        blitzortung.set_log_level(logging.WARNING)

    # Use lock unless disabled
    lock_context = nullcontext() if options.no_lock else LockWithTimeout('/tmp/.bo-import2.lock').locked(10)

    try:
        with lock_context:
            count = update_strikes(hours=options.hours)
            logger.info("Import completed: %d new strikes inserted", count)
            return 0

    except FailedToAcquireException:
        logger.warning("Could not acquire lock - another import may be running")
        return 1
    except Exception as e:
        logger.error("Import failed: %s", e, exc_info=options.debug)
        return 1


if __name__ == "__main__":
    import sys
    sys.exit(main())



# -*- coding: utf8 -*-

"""

   Copyright 2014-2016 Andreas Würl

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""

import datetime

import shapely.geometry.base
import shapely.wkb

import psycopg2

class BaseInterval:
    """
    Basic interval range
    """

    __slots__ = ['__start', '__end']

    def __init__(self, start=None, end=None):
        self.__start = start
        self.__end = end

    @property
    def start(self):
        return self.__start

    @property
    def end(self):
        return self.__end

    def __str__(self):
        return '[' + (str(self.start) if self.start else '') + ' : ' + (str(self.end) if self.end else '') + ']'


class IdInterval(BaseInterval):
    """
    Interval of Id
    """

    __slots__ = []

    def __init__(self, start=None, end=None):
        if start and not isinstance(start, int):
            raise ValueError("start should be an integer value")
        if end and not isinstance(end, int):
            raise ValueError("end should be an integer value")

        super().__init__(start, end)


class TimeInterval(BaseInterval):
    """
    Time interval
    """

    __slots__ = []

    def __init__(self, start=None, end=None):
        if start and not isinstance(start, datetime.datetime):
            raise ValueError("start should be a datetime value")
        if end and not isinstance(end, datetime.datetime):
            raise ValueError("end should be a datetime value")

        super().__init__(start, end)

    @property
    def duration(self):
        return self.end - self.start

    def contains(self, timestamp):
        return self.start <= timestamp < self.end

    def minutes(self) -> int:
        if self.start and self.end:
            return int((self.end - self.start).total_seconds() // 60)
        else:
            raise ValueError("incomplete time interval")


class Query:
    """
    simple class for building of complex queries
    """

    __slots__ = ['conditions', 'groups', 'groups_having', 'parameters', 'limit', 'order', 'default_conditions']

    def __init__(self):
        self.conditions = []
        self.groups = []
        self.groups_having = []
        self.parameters = {}
        self.limit = None
        self.order = []
        self.default_conditions = {
            'time_interval': self.add_time_interval,
            'id_interval': self.add_id_interval,
            'geometry': self.add_geometry,
            'order': self._set_order,
            'limit': self.set_limit
        }

    def add_group_by(self, group_by):
        self.groups.append(group_by)
        return self

    def _set_order(self, order):
        order_items = order if type(order) is list else [order]
        return self.set_order(*order_items)

    def set_order(self, *order_items):
        if len(self.order) > 0:
            raise RuntimeError("overriding Query.limit")

        self.order = [(order_item if isinstance(order_item, Order) else Order(order_item)) for order_item in
                      order_items]
        return self

    def set_limit(self, limit):
        if self.limit:
            raise RuntimeError("overriding Query.limit")
        self.limit = limit
        return self

    def add_condition(self, condition, **parameters):
        self.conditions.append(condition)
        self.add_parameters(**parameters)
        return self

    def add_group_having(self, condition, **parameters):
        self.groups_having.append(condition)
        self.add_parameters(**parameters)
        return self

    def add_parameters(self, **parameters):
        self.parameters.update(parameters)
        return self

    def __str__(self):
        sql = ""

        if self.conditions:
            sql += 'WHERE ' + ' AND '.join(self.conditions) + ' '

        if self.groups:
            sql += 'GROUP BY ' + ', '.join(self.groups) + ' '

            if self.groups_having:
                sql += 'HAVING ' + ' AND '.join(self.groups_having) + ' '

        if self.order:
            def build_order_query(order): return order.get_column() + (' DESC' if order.is_desc() else '')

            order_query_elements = map(build_order_query, self.order)
            sql += 'ORDER BY ' + ', '.join(order_query_elements) + ' '

        if self.limit:
            sql += 'LIMIT ' + str(self.limit) + ' '

        return sql.strip()

    def get_parameters(self):
        return self.parameters

    def set_default_conditions(self, **kwargs):
        for keyword, value in kwargs.items():
            if keyword in self.default_conditions and value:
                self.default_conditions[keyword](value)
        return self

    def add_time_interval(self, time_interval):
        if time_interval.start:
            self.add_condition('"timestamp" >= %(start_time)s', start_time=time_interval.start)

        if time_interval.end:
            self.add_condition('"timestamp" < %(end_time)s', end_time=time_interval.end)

    def add_id_interval(self, id_interval):
        if id_interval.start:
            self.add_condition('id >= %(start_id)s', start_id=id_interval.start)

        if id_interval.end:
            self.add_condition('id < %(end_id)s', end_id=id_interval.end)

    def add_geometry(self, geometry):
        if geometry.is_valid:
            self.add_condition('ST_GeomFromWKB(%(envelope)s, %(srid)s) && geog',
                               envelope=psycopg2.Binary(shapely.wkb.dumps(geometry.envelope)))

            if not geometry.equals(geometry.envelope):
                self.add_condition(
                    'ST_Intersects(ST_GeomFromWKB(%(geometry)s, %(srid)s), ' +
                    'ST_Transform(geog::geometry, %(srid)s))',
                    geometry=psycopg2.Binary(shapely.wkb.dumps(geometry)))

        else:
            raise ValueError("invalid geometry in db.Strike.select()")


class SelectQuery(Query):
    __slots__ = ['table_name', 'columns']

    def __init__(self):
        super().__init__()
        self.table_name = ""
        self.columns = []

    def set_table_name(self, table_name):
        self.table_name = table_name
        return self

    def set_columns(self, *columns):
        self.columns = columns
        return self

    def add_column(self, column):
        self.columns.append(column)
        return self

    def __str__(self):
        sql = 'SELECT '

        if self.columns:
            sql += ', '.join(self.columns) + ' '

        sql += 'FROM ' + self.table_name + ' '

        sql += super().__str__()

        return sql.strip()


class GridQuery(SelectQuery):
    __slots__ = ['grid']

    def __init__(self, grid, count_threshold=0):
        super().__init__()

        self.grid = grid

        self.add_parameters(
            srid=grid.srid,
            xmin=grid.x_min,
            xdiv=grid.x_div,
            ymin=grid.y_min,
            ydiv=grid.y_div,
        )

        self.set_columns(
            'TRUNC((ST_X(ST_Transform(geog::geometry, %(srid)s)) - %(xmin)s) / %(xdiv)s)::integer AS rx',
            'TRUNC((ST_Y(ST_Transform(geog::geometry, %(srid)s)) - %(ymin)s) / %(ydiv)s)::integer AS ry',
            'count(*) AS strike_count',
            'max("timestamp") as "timestamp"'
        )

        env = self.grid.env

        if env.is_valid:
            self.add_condition('ST_GeomFromWKB(%(envelope)s, %(envelope_srid)s) && geog',
                               envelope=psycopg2.Binary(shapely.wkb.dumps(env)),
                               envelope_srid=grid.srid)
        else:
            raise ValueError("invalid Raster geometry in db.query.GridQuery.__init__()")

        self.add_group_by('rx')
        self.add_group_by('ry')

        if count_threshold > 0:
            self.add_group_having("count(*) > %(count_threshold)s", count_threshold=count_threshold)


class GlobalGridQuery(SelectQuery):
    __slots__ = ['grid']

    def __init__(self, grid, count_threshold=0):
        super().__init__()

        self.grid = grid

        self.add_parameters(
            srid=grid.srid,
            xdiv=grid.x_div,
            ydiv=grid.y_div,
        )

        self.set_columns(
            'ROUND((ST_X(ST_Transform(geog::geometry, %(srid)s)) - %(xdiv)s * 0.5) / %(xdiv)s)::integer AS rx',
            'ROUND((ST_Y(ST_Transform(geog::geometry, %(srid)s)) - %(ydiv)s * 0.5) / %(ydiv)s)::integer AS ry',
            'count(*) AS strike_count',
            'max("timestamp") as "timestamp"'
        )

        self.add_group_by('rx')
        self.add_group_by('ry')

        if count_threshold > 0:
            self.add_group_having("count(*) > %(count_threshold)s", count_threshold=count_threshold)


class Order:
    """
    definition for query search order
    """

    __slots__ = ['column', 'desc']

    def __init__(self, column, desc=False):
        self.column = column
        self.desc = desc

    def get_column(self):
        return self.column

    def is_desc(self):
        return self.desc


class Center:
    """
    definition of query center point
    """

    __slots__ = ['point']

    def __init__(self, center):
        self.point = center

    def get_point(self):
        return self.point



# -*- coding: utf8 -*-

"""

   Copyright 2014-2016 Andreas Würl

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""

import configparser
import os
from typing import Optional

from injector import Module, singleton, inject, provider


@singleton
class Config:
    @inject
    def __init__(self, config_parser: configparser.ConfigParser):
        self.config_parser = config_parser

    def get_username(self):
        return self.config_parser.get('auth', 'username')

    def get_password(self):
        return self.config_parser.get('auth', 'password')

    def get_db_connection_string(self):
        host = self.config_parser.get('db', 'host')
        port = self.config_parser.get('db', 'port', fallback='5432')
        dbname = self.config_parser.get('db', 'dbname')
        username = self.config_parser.get('db', 'username')
        password = self.config_parser.get('db', 'password')

        return "host='%s' port=%s dbname='%s' user='%s' password='%s'" % (host, port, dbname, username, password)

    def get_webservice_port(self):
        return int(self.config_parser.get('webservice', 'port'))

    def __str__(self):
        return "Config(user: %s, pass: %s)" % (self.get_username(), len(self.get_password()) * '*')


def config():
    from blitzortung import INJECTOR

    return INJECTOR.get(Config)


class ConfigModule(Module):
    @singleton
    @provider
    def provide_config_parser(self) -> configparser.ConfigParser:
        config_file_path = self.find_config_file_path()

        if config_file_path is None:
            raise ValueError("No configuration file found")

        config_parser = configparser.ConfigParser()
        config_parser.read(config_file_path)
        return config_parser

    def find_config_file_path(self) -> Optional[str]:
        config_file_name = "blitzortung.conf"
        for config_dir_name in [".", "/etc/"]:
            config_file_path = os.path.join(config_dir_name, config_file_name)
            if os.path.exists(config_file_path):
                return config_file_path



# -*- coding: utf8 -*-

"""

   Copyright 2014-2022 Andreas Würl

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""

import logging

__version__ = '1.7.1'


# -----------------------------------------------------------------------------
# Constants.
# -----------------------------------------------------------------------------

# -----------------------------------------------------------------------------
#   Custom exceptions.
# -----------------------------------------------------------------------------


class Error(Exception):
    """
    General Blitzortung error class.
    """
    pass


# -----------------------------------------------------------------------------
#   Public interface and exports.
# -----------------------------------------------------------------------------


import injector

from . import builder
from . import config
from . import data
from . import dataimport
from . import db
from . import geom
from . import util
from . import base

INJECTOR = injector.Injector(
    [config.ConfigModule(), db.DbModule()])

root_logger = logging.getLogger(__name__)
root_logger.setLevel(logging.WARN)


def set_parent_logger(logger):
    logger.parent = root_logger


def set_log_level(log_level):
    root_logger.setLevel(log_level)


def add_log_handler(log_handler):
    root_logger.addHandler(log_handler)



# -*- coding: utf8 -*-

"""

   Copyright 2014-2016 Andreas Würl

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""

import math

import pyproj


class EqualityAndHash:
    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.__dict__ == other.__dict__
        return NotImplemented

    def __ne__(self, other):
        if isinstance(other, self.__class__):
            return not self.__eq__(other)
        return NotImplemented

    def __hash__(self):
        return hash(tuple(sorted(self.__dict__.items())))


class Point:
    """
    Base class for Point like objects
    """

    __radians_factor = math.pi / 180

    __geod = pyproj.Geod(ellps='WGS84', units='m')

    __slots__ = ('x', 'y')

    def __init__(self, x_coord_or_point, y_coord=None):
        (self.x, self.y) = self.__get_point_coordinates(x_coord_or_point, y_coord)

    def distance_to(self, other):
        return self.geodesic_relation_to(other)[1]

    def azimuth_to(self, other):
        return self.geodesic_relation_to(other)[0]

    def geodesic_shift(self, azimuth, distance):
        result = self.__geod.fwd(self.x, self.y, azimuth / self.__radians_factor, distance, radians=False)
        return Point(result[0], result[1])

    def geodesic_relation_to(self, other):
        result = self.__geod.inv(self.x, self.y, other.x, other.y, radians=False)
        return result[0] * self.__radians_factor, result[2]

    @staticmethod
    def __get_point_coordinates(x_coord_or_point, y_coord):
        if isinstance(x_coord_or_point, Point):
            return x_coord_or_point.x, x_coord_or_point.y
        else:
            return x_coord_or_point, y_coord

    def __eq__(self, other):
        return self.equal(self.x, other.x) and self.equal(self.y, other.y)

    @staticmethod
    def equal(a, b):
        return abs(a - b) < 1e-4

    def __str__(self):
        return "(%.4f, %.4f)" % (self.x, self.y)

    def __hash__(self):
        return hash(self.x) ^ hash(self.y)



def decode(data):
    h = 256
    e = {}

    c = data[0]
    f = c
    g = c
    o = h
    for b, char in enumerate(data[1:]):
        a = ord(char)
        a = char if h > a else e.get(a, f + c)
        g += a

        c = a[0]
        e[o] = f + c
        o += 1
        f = a
    return g



from contextlib import contextmanager

from fasteners import InterProcessLock


# Taken from https://github.com/harlowja/fasteners/issues/51#issuecomment-755723163
class FailedToAcquireException(Exception):
    pass


class LockWithTimeout(InterProcessLock):

    @contextmanager
    def locked(self, timeout):
        ok = self.acquire(timeout=timeout)
        if not ok:
            raise FailedToAcquireException()
        try:
            yield
        finally:
            self.release()
